--
-- Handles initialisation, versioning and migration of saved variables
--

local revision = tonumber(string.sub("$Revision: 66 $", 12, -3))
local rss = _G.RShamanStat
if rss.revision < revision then rss.revision = revision end

-- Archive version and code version are separate, as not every code change changes the archive
local archiveVersion = 29


-- Instructions for changing the archive:
-- 1. Increase the archiveVersion
-- 2. Write a function to migrate data from the previous version to the next
-- 3. Add the function with its version number to the 2-dimensional migration table
-- 4. Update the initial archive initialisation routine
-- 5. If data cannot be migrated, delete all migration functions and add ArchiveClear with the new Version
--    to the migration table

-- Initialize the archive
-- returns either a new or a modified version of the passed archive
function rss.LoadArchive(archive)
    if archive == nil then
        archive = rss.CreateArchive(archive)
    end

    print(("RSS: Loading Archive v%d"):format(archive.version))

    if archive.version == nil then
        archive.version = 0
    end
    return archive
end

-- Migration helper function to clear all saved records if archive cannot be migrated or does not match current haste breakpoints
local ArchiveClear = function(archive)
	for i in ipairs(archive) do
		archive[i] = nil;
	end
end

-- Migrate or clear the archive
-- Unlike the load function, this will only happen when in Resto spec
function rss.ValidateArchive(archive)
	if not rss.playerState or rss.playerState.role ~= "HEALER" then return; end
    if archive.version < archiveVersion then
        -- prevent data loss by testing the code prior to running the real migration
        -- TODO: save the old archive under archive.old and provide a way to revert the data
        local proceed = rss.TestMigration();
        if not proceed then
            print("RSS: Migrations are currently not working. Aborting...")
			return;
        end
		print(("RSS: Migrating Archived Records from v%d to v%d"):format(archive.version, archiveVersion))
		rss.MigrateArchive(archive)
    end
	-- if race and/or talents have changed in a way that affects spell haste, clear all data and start over
	if archive.ancestralSwiftness ~= rss.playerState.ancestralSwiftnessActive or
	   archive.isGoblin ~= rss.playerState.isGoblin or 
	   not archive.hasteBreakpoints then
		if archive.isGoblin ~= rss.playerState.isGoblin then
			print("RSS: Race changed to or from Goblin!");
		end
		if archive.ancestralSwiftness ~= rss.playerState.ancestralSwiftnessActive then
			print("RSS: Ancestral Swiftness talent changed!");
		end
		if not archive.hasteBreakpoints then
			print("RSS: Haste breakpoints missing (error)");
		end
		ArchiveClear(archive);
		archive.ancestralSwiftness = rss.playerState.ancestralSwiftnessActive;
		archive.isGoblin = rss.playerState.isGoblin;
		archive.hasteBreakpoints = rss.playerState.hasteBreakpoints;
	end
end

-- returns a new empty archive
function rss.CreateArchive()
    archive = { };
    archive.version = archiveVersion;
    archive.filter = {
        difficulty = 0,     -- difficulty 0 = any difficulty; otherwise archive.filter.difficulty = fight.difficulty
        useAll     = true,  -- if false, /rss sum will only show boss fights
    }
	archive.isGoblin = rss.playerState and rss.playerState.isGoblin;
	if rss.playerState and rss.playerState.role == "HEALER" then
		archive.ancestralSwiftness = rss.playerState.ancestralSwiftnessActive;
		archive.hasteBreakpoints = rss.playerState.hasteBreakpoints;
	else  -- will be updated when we change back to Resto
		archive.ancestralSwiftness = nil;
		archive.hasteBreakpoints = nil;
	end
    return archive;
end


function rss.CreateRecord()
	local record = 
        {
            startTime               =                        0,   -- time counter for start of current combat
            totalTime               =                       .0,   -- aggregator for total time in combat

            effectiveTotal          =                       .0,   -- total effective healing in current combat
			leakage					=					   { },   -- effective healing for which we could not analyze stats
            criticalTotal           =                       .0,   -- effective healing from Critical in current combat
            masteryTotal            =                       .0,   -- effective healing from Mastery in current combat
			incrementalSpellpower   =                       .0,   -- increase to effective healing from +1 Spellpower
			incrementalCritical     =               		.0,   -- increase to effective healing from +1 Critical Rating
            incrementalMastery      =                       .0,   -- increase to effective healing from +1 Mastery Rating
			incrementalHaste        =                      { },   -- increase to effective healing from +1 Haste Rating by changing from baseHasteRating to each breakpoint
            weightedTotal           =                       .0,   -- total weighted by the mastery effect of each heal
			effectiveSubTotal       =                       .0,   -- total effective healing of spells where we know the mastery effect (i.e. not RM/AG/SLT)
            spells                  =                       {},   -- a table of spellIDs and how often they healed somebody
            resurgence              =                       .0,   -- sum of mana gained from resurgence

            crit                    =    GetSpellCritChance(rss.SCHOOL_NATURE),   -- critical strike chance
            haste                   = UnitSpellHaste("player"),   -- haste
            mastery                 =          GetMastery()*3.,   -- mastery
            spellpower              =   GetSpellBonusHealing(),   -- spellpower
			baseHasteRating         =   rss.playerState.baseHasteRating,

            difficulty              = select(3, GetInstanceInfo()),  --  0 - None; 1 - 5man; 2 = 5HC; 3 = 10N; 4 = 25N; 5 = 10H; 6 = 25H; 7 = LFR; 8 = CM; 9 = 40N; 11 = HScen; 12 = NScen
            instance                =        GetInstanceInfo(),   -- localized instance name
            isBoss                  =                    false,   -- flag to indicate if a boss encounter is in progress
            encounterName           =                "Unknown",   -- highest classification enemy detected
        };
	for i in ipairs(rss.playerState.hasteBreakpoints) do
		record.incrementalHaste[i] = 0.0;
	end
	record.incrementalHaste[0] = 0.0;
	return record;
end

function rss.CreateAggregateRecord(name, archive, filter)
    local a = rss.CreateRecord()
    local recordCount = 0;
    a.encounterName = name
    a.difficulty = -1;
    a.spellpower = 0;
    a.haste = 0;
    a.mastery = 0;
    a.crit = 0;
	a.baseHasteRating = rss.playerState.baseHasteRating;
    if (#archive > 0) then
        for i, r in ipairs(archive) do
            if (filter(r)) then
                recordCount = recordCount + 1;
                if a.startTime <= 0 or a.startTime > r.startTime then
                    a.startTime = r.startTime
                end

                if a.difficulty == -1 then
                    a.difficulty = r.difficulty;
                elseif a.difficulty ~= r.difficulty then
                    a.difficulty = 0;
                end


                a.effectiveTotal = a.effectiveTotal + r.effectiveTotal
				for reason, leak in pairs(r.leakage) do
					if not a.leakage[reason] then
						a.leakage[reason] = 0.0
					end
					a.leakage[reason] = a.leakage[reason] + leak
				end
				a.effectiveSubTotal = a.effectiveSubTotal + r.effectiveSubTotal
                a.totalTime = a.totalTime + r.totalTime
                a.criticalTotal = a.criticalTotal + r.criticalTotal
                a.masteryTotal = a.masteryTotal + r.masteryTotal
				a.incrementalSpellpower = a.incrementalSpellpower + r.incrementalSpellpower
				a.incrementalCritical = a.incrementalCritical + r.incrementalCritical
                a.incrementalMastery = a.incrementalMastery + r.incrementalMastery
				for i, b in ipairs(archive.hasteBreakpoints) do
					a.incrementalHaste[i] = a.incrementalHaste[i] + 
						r.incrementalHaste[i] * (b.rating - r.baseHasteRating) -
						r.incrementalHaste[0] * (a.baseHasteRating - r.baseHasteRating)
				end
				a.incrementalHaste[0] = a.incrementalHaste[0] + r.incrementalHaste[0];
                a.isBoss = r.isBoss
                a.weightedTotal = a.weightedTotal + r.weightedTotal

                a.spellpower = a.spellpower + r.spellpower
                a.haste = a.haste + r.haste
                a.crit = a.crit + r.crit
                a.mastery = a.mastery + r.mastery

                -- sum spell counts
                for spellid, count in pairs(r.spells) do
                    if a.spells[spellid] then
                        a.spells[spellid] = count + a.spells[spellid]
                    else
                        a.spells[spellid] = count
                    end
                end

            end
        end
		for i, b in ipairs(archive.hasteBreakpoints) do
			a.incrementalHaste[i] = a.incrementalHaste[i]/(b.rating - a.baseHasteRating);
		end
    end

    if recordCount == 0 then
        a.spellpower = 0;
        a.haste = 0;
        a.mastery = 0;
        a.crit = 0;
    else
        a.spellpower = a.spellpower / recordCount;
        a.haste = a.haste / recordCount;
        a.mastery = a.mastery / recordCount;
        a.crit = a.crit / recordCount;
    end

    return a
end

function rss.SaveRecord(archive, record)

    table.insert(archive, record)
end

function rss.ResurgenceGains(record)
    local sum = 0
	for spellid, casts in pairs(record.spells) do
		if rss.SPELLS[spellid] and rss.SPELLS[spellid].resurgence then
			sum = sum + casts * rss.SPELLS[spellid].resurgence
		end
	end
    return sum
end

function rss.AnalyzeRecord(record)
    local r = record;

    -- initialize and document all values, so we can leave things we can't calculate for a specific record unset
    local a = {
        hps = 0.0,
        leakageTotal = 0.0,
        analyzedTotal = 0.0,
        -- average health ratio (0.0 - 1.0)
        averageHealth = 0.0,
        -- total gain in percent
        masteryTotalThroughputGain = 0.0,
        critTotalThroughputGain = 0.0,
        -- rating required for 1% throughput
        spellpowerInc = 0.0,
        intellectInc = 0.0,
        masteryInc = 0.0,
        critInc = 0.0,
		hasteInc = 0.0,
		hasteNoBrkInc = 0.0,
		hasteBrk = 0.0,
        -- same as above, but for 1% more regen
        critRegenInc = 0.0,
        spiritRegenInc = 0.0,

        -- average expected mp5 for the entire fight
        crit_mp5 = 0.0,
        spirit_mp5 = 338.61, -- hardcoded gain without MTT
        -- potential mana gain from using glyph of water shield, may be negative.
        waterShieldGlyphManaGain = 0.0,

    };

    if record.totalTime > 0 then
        a.hps = r.effectiveTotal / r.totalTime;
    end

    local leakageTotal = 0.0
    for reason, leaks in pairs(record.leakage) do
        leakageTotal = leakageTotal + leaks;
    end
    a.leakageTotal = leakageTotal;

    local analyzedTotal = math.max(0.0, record.effectiveTotal - leakageTotal)
    a.analyzedTotal = analyzedTotal
    if analyzedTotal > 0 then
        local critPct = 100 * record.criticalTotal / analyzedTotal
        local mstPct = 100 * record.masteryTotal / analyzedTotal
        local critInc = nil
        local mstInc = nil
		local hstInc = nil
		local hstBrk = nil
		local hstNoBrkInc = nil
        local spInc = nil
        local intInc = nil

        if record.incrementalCritical > 0 then
            critInc = 0.01 * analyzedTotal / record.incrementalCritical
        end

        -- the factor 1.05 reflects the +5% stat buff to Intellect in raids
        if record.incrementalSpellpower > 0 then
            spInc = 0.01 / (record.incrementalSpellpower / analyzedTotal);
            if critInc then
                intInc = 0.01 / (1.05*(0.01/spInc) + (rss.RATING_TO_CRIT/rss.INT_TO_CRIT)*(0.01/critInc))
            end
        end

        if record.incrementalMastery > 0 then
            mstInc = 0.01 / (record.incrementalMastery / analyzedTotal)
        end
		
		local selectedHasteBreakpointIndex = nil;
		local selectedHasteBreakpoint = nil;
		for i, breakpoint in ipairs(rss.playerState.hasteBreakpoints) do
			if breakpoint.rating > record.baseHasteRating then
				selectedHasteBreakpointIndex = i;
				selectedHasteBreakpoint = breakpoint.rating;
				break;
			end
		end
		
		if record.incrementalHaste[selectedHasteBreakpointIndex] ~= 0 then
			hstInc = 0.01 / (record.incrementalHaste[selectedHasteBreakpointIndex] / analyzedTotal)
			hstBrk = selectedHasteBreakpoint;
		end
		if record.incrementalHaste[0] ~= 0 then
			hstNoBrkInc = 0.01 / (record.incrementalHaste[0] /analyzedTotal)
		end

        -- mana gain per 5 seconds from 600 spirit
        a.spirit_mp5 = 1.1287 * 600 * 0.5
        -- mana gain per 5 seconds from 600 (1%) of crit
        a.crit_mp5 = rss.ResurgenceGains(record) * 5.0 / 100.0 / record.totalTime
        -- current combat regen
        local _, mp1Combat = GetManaRegen();
        a.spiritRegenInc = 600*(0.01 * 5 * mp1Combat)/a.spirit_mp5;
        if a.crit_mp5 > 0 then
            a.critRegenInc = 600*(0.01 * 5 * mp1Combat)/a.crit_mp5;
        else
            a.critRegenInc = 0;
        end

        a.critTotalThroughputGain = critPct;
        a.masteryTotalThroughputGain = mstPct;

        a.spellpowerInc = spInc
        a.intellectInc = intInc
        a.critInc = critInc
        a.masteryInc = mstInc
		a.hasteInc = hstInc
		a.hasteNoBrkInc = hstNoBrkInc
		a.hasteBrk = hstBrk
    end
    if r.effectiveSubTotal > 0 then
        a.averageHealth = 1 - r.weightedTotal/r.effectiveSubTotal
    end

    local ws_procs = (r.spells[rss.PROC_WATER_SHIELD] or 0)
    local ws_unglyphed = r.totalTime / 5.0 * rss.WATER_SHIELD.mp5 + ws_procs * rss.WATER_SHIELD.proc
    local ws_glyphed = r.totalTime / 5.0 * rss.WATER_SHIELD.glyphed_mp5 + ws_procs * rss.WATER_SHIELD.glyphed_proc
    a.waterShieldGlyphManaGain = ws_glyphed - ws_unglyphed


    return a;
end

function rss.RecordToString(record)
	local s ="";
    local healing = record.effectiveTotal / 1000.
    if record.totalTime > 0 then
        local hps = healing / record.totalTime
        s = s..("RSS: Effective Healing Done: %.1fk ; Effective HPS: %.1fk\n"):format(healing,hps)
    else
        s = s..("RSS: Effective Healing Done: %.1fk\n"):format(healing)
    end
	local leakageTotal = 0.0
	for reason, leaks in pairs(record.leakage) do
		leakageTotal = leakageTotal + leaks;
	end
	
	local analyzedTotal = math.max(0.0, record.effectiveTotal - leakageTotal)
    if analyzedTotal > 0 then
        local critPct = 100 * record.criticalTotal / analyzedTotal
        local mstPct = 100 * record.masteryTotal / analyzedTotal
		local critInc = nil
		local mstInc = nil
		local hstInc = nil
		local hstBrk = nil
		local hstNoBrk = nil
		local spInc = nil
		local intInc = nil
				
		if record.incrementalCritical > 0 then
			critInc = 0.01 * analyzedTotal / record.incrementalCritical
		end

		-- the factor 1.05 reflects the +5% stat buff to Intellect in raids
		if record.incrementalSpellpower > 0 then
			spInc = 0.01 / (record.incrementalSpellpower / analyzedTotal);
			if critInc then
				intInc = 0.01 / (1.05*(0.01/spInc) + (rss.RATING_TO_CRIT/rss.INT_TO_CRIT)*(0.01/critInc))
			end
		end
		
        if record.incrementalMastery > 0 then
			mstInc = 0.01 / (record.incrementalMastery / analyzedTotal)
		end
		
		local selectedHasteBreakpointIndex = nil;
		local selectedHasteBreakpoint = nil;
		for i, breakpoint in ipairs(rss.playerState.hasteBreakpoints) do
			if breakpoint.rating > record.baseHasteRating then
				selectedHasteBreakpointIndex = i;
				selectedHasteBreakpoint = breakpoint.rating;
				break;
			end
		end
		
		if record.incrementalHaste[selectedHasteBreakpointIndex] ~= 0 then
			hstInc = 0.01 / (record.incrementalHaste[selectedHasteBreakpointIndex] / analyzedTotal)
			hstBrk = selectedHasteBreakpoint;
		end
		if record.incrementalHaste[0] ~= 0 then
			hstNoBrkInc = 0.01 / (record.incrementalHaste[0] / analyzedTotal)
		end
		
        -- mana gain per 5 seconds from 600 spirit
        local spirit_mp5 = 1.1287 * 600 * 0.5
        -- mana gain per 5 seconds from 600 (1%) of crit
        local crit_mp5 = rss.ResurgenceGains(record) * 5.0 / 100.0 / record.totalTime
		-- current combat regen
		local _, mp1Combat = GetManaRegen();
		local spiritRegenInc = 600*(0.01 * 5 * mp1Combat)/spirit_mp5;
		local critRegenInc = nil;
		if crit_mp5 > 0 then
			critRegenInc = 600*(0.01 * 5 * mp1Combat)/crit_mp5;
		end

        s = s..("RSS: Throughput from Critical: %.2f%%\n"):format(critPct);
        s = s..("RSS: Throughput from Mastery:  %.2f%%\n"):format(mstPct);
		
		if spInc then
			s = s..("RSS: Spell Power for +1%% Throughput: %.1f\n"):format(spInc);
		end
		if intInc then
			s = s..("RSS: Intellect for +1%% Throughput: %.1f\n"):format(intInc);
		end
        if critInc then
            s = s..("RSS: Critical Rating for +1%% Throughput: %.1f\n"):format(critInc);
        end
        if mstInc then
            s = s..("RSS: Mastery Rating for +1%% Throughput: %.1f\n"):format(mstInc);
        end
        if hstInc then
            s = s..("RSS: Haste Rating for +1%% Throughput (to %d breakpoint): %.1f\n"):format(hstBrk, hstInc);
        end
		if hstNoBrkInc then
			s = s..("RSS: Haste Rating for +1%% Throughput (< %d breakpoint): %.1f\n"):format(hstBrk, hstNoBrkInc);
		end
		if critRegenInc then
			s = s..("RSS: Spirit/Crit Rating for +1%% mp5: %.1f / %.1f"):format(spiritRegenInc, critRegenInc);
		end
    end
	return s;
end

function rss.PrintRecord(record)
-- when printing a string with embedded \n, lua seems to indent each line after the first
-- so it is slightly better looking to have the header be the first line, so the following
-- lines are all indented the same
-- another side effect of doing it as one long string with \n's is that you can't scroll through 
-- the printout one line at a time, because the chat window thinks it's all one line
    print(("---- %s ----\n"):format(record.encounterName)..rss.RecordToString(record))
end

function rss.PrintRSM()
    if (#RShamanStatArchive == 0) then
        print("RSS: no data available");
        return;
    end

    for i, v in ipairs(RShamanStatArchive) do
        if (RShamanStatArchive.filter.useAll or v.isBoss) then
            if v.effectiveTotal ~= 0 and v.effectiveSubTotal ~= 0 then
                print(("RSS: #%02d - %s: Healing Done: %.1fk, Mastery Effect: %.2f"):format(i, v.encounterName, v.effectiveTotal/1000, v.weightedTotal/v.effectiveSubTotal));
            else
                print(("RSS: #%02d - %s: No Healing Done"):format(i, v.encounterName));
            end
        end
    end
end

-- Migration Helper function that adds unset keys to a certain default value
-- keys should be formatted as {"newKey1" = .0, "newKey2" = "DefVal", .. }
local AddKeys = function(archive, keys)
    for i, record in ipairs(archive) do
        for key, value in pairs(keys) do
            if (not archive[i][key]) then
                archive[i][key] = value
            end
        end
    end
end

-- Migration Helper function that will rename a key
-- keys should be formatted as {"newKey1 = "oldKey1, "newKey2" = "oldKey2", .. }
local RenameKeys = function(archive, keys)
    for i, record in ipairs(archive) do
        for new, old in pairs(keys) do
            archive[i][new] = record[old]
            archive[i][old] = nil
        end
    end
end

-- Migration helper function that passes every record to a function
-- the passed function needs to take a 'record' as it's single parameter
local Map = function(archive, fn)
    for i, record in ipairs(archive) do
        fn(archive[i])
    end
end

local ArchiveMigration_29 = function(archive)
	ArchiveClear(archive);
	if rss.playerState then
		archive.isGoblin = rss.playerState.isGoblin;
		archive.ancestralSwiftness = rss.playerState.ancestralSwiftnessActive;
		archive.hasteBreakpoints = rss.playerState.hasteBreakpoints;
	end
end

-- Don't use keyword indices here to ensure a fixed order when we iterate over it.
local migrations = {
	{v=29, f=ArchiveMigration_29},
}

-- modifies the passed archive to be compatible with the current version
rss.MigrateArchive = function(archive)

    for i, m in ipairs(migrations) do
        if archive.version < m.v then
            m.f(archive)
            -- update the version after every step, in case the next migration fails
            archive.version = m.v
        end
    end
end


-- A really simple test to see if the migrations are working as intended
rss.TestMigration = function()

	local testArchive = { version = 29 };

	if testArchive.version == archiveVersion then return true; end
	
    rss.MigrateArchive(testArchive)

    if not testArchive then return false; end
    if not testArchive.version == archiveVersion then return false; end
    local r = rss.CreateAggregateRecord("Test", testArchive, function(r) return true; end)
    if not r then return false; end
    if not (r.effectiveTotal > 0.) then return false; end;
    return true;
end
