local revision = tonumber(string.sub("$Revision: 67 $", 12, -3))
local rss = _G.RShamanStat
if rss.revision < revision then rss.revision = revision end

-- Crit-related
function rss.EvaluateCriticalRating(amount, deficit, critical)	
	local m = rss.playerState.criticalMultiplier;
	local criticalContribution, criticalIncrement;
	if critical then
		criticalContribution = math.min(amount, deficit) - math.min(amount/m,deficit);
		criticalIncrement = criticalContribution;
	else
		criticalContribution = 0.0;
		criticalIncrement = math.min(amount*m,deficit) - math.min(amount,deficit);
	end
	criticalIncrement = criticalIncrement * (0.01/rss.RATING_TO_CRIT);
	return criticalIncrement, criticalContribution;
end

function rss.EvaluateAncestralAwakening(amount, deficit, critical)
	local probSetBonus = rss.playerState.aaSetBonusT15 and rss.SET_BONUS_T15_CHANCE or 0.0;
	local m = rss.playerState.criticalMultiplier;
	local criticalContribution, criticalIncrement;
	if critical then -- amount is the size of the actual (crit) AA
		criticalContribution = math.min(amount, deficit) 
			- probSetBonus * math.min(rss.SET_BONUS_T15_RATIO*amount/m,deficit);
		criticalIncrement = criticalContribution;
	else -- amount is the size of the original (non-crit) heal
		criticalContribution = 0.0;
		criticalIncrement = math.min(rss.SPELLS[rss.ANCESTRAL_AWAKENING].copy * amount * m, deficit)
			- probSetBonus * math.min(rss.SET_BONUS_T15_FRACTION * amount, deficit);
	end
	criticalIncrement = criticalIncrement * (0.01/rss.RATING_TO_CRIT);
	return criticalIncrement, criticalContribution;
end

-- called only for set bonus procs (non-crit) where we originally guessed the target of the AA, but now know it
function rss.CorrectAncestralAwakening(origAmount, assumedDeficit, aaAmount, actualDeficit)
	local probSetBonus = rss.playerState.aaSetBonusT15 and rss.SET_BONUS_T15_CHANCE or 0.0;
	local m = rss.playerState.criticalMultiplier;
	local criticalIncrement = (math.min(aaAmount*m/rss.SET_BONUS_T15_RATIO, actualDeficit) 
								- probSetBonus*math.min(aaAmount, actualDeficit))
							- (math.min(rss.SPELLS[rss.ANCESTRAL_AWAKENING].copy * origAmount * m, assumedDeficit) 
								- probSetBonus*math.min(rss.SET_BONUS_T15_FRACTION*origAmount, assumedDeficit));
	criticalIncrement = criticalIncrement * (0.01/rss.RATING_TO_CRIT);
	return criticalIncrement;
end

-- special case for Glyph of Healing Wave, since both the original heal and the glyph heal (or both) can crit
function rss.GlyphOfHealingWaveCritical(origAmount, amount, deficit, origCritical)
	local m = rss.playerState.criticalMultiplier;
	local f = rss.SPELLS[rss.GLYPH_OF_HEALING_WAVE].copy;
	local baseHW = origCritical and origAmount/m or origAmount;
	local c = rss.playerState.criticalChance[rss.SCHOOL_NATURE];
	local criticalContribution = math.max(0, math.min(amount, deficit)-math.min(f*baseHW, deficit));
	local criticalIncrement = 2*(-(1-c)*math.min(f*baseHW,deficit)+(1-2*c)*math.min(f*baseHW*m,deficit)+c*math.min(f*baseHW*m*m,deficit));
	criticalIncrement = criticalIncrement * (0.01/rss.RATING_TO_CRIT);
	return criticalIncrement, criticalContribution;
end

-- Haste-related

-- Estimate the change in periodic healing of a cast, based on the difference between all hypothetical haste breakpoints and the actual haste value
function rss.VaryTicks(cast)
	if not cast.events or cast.healing <= 0 or cast.spellID == rss.ASCENDANCE or cast.spellID == rss.AURA_ANCESTRAL_GUIDANCE then return; end
	cast.deltaTicks = { };

	local castHaste = cast.haste;
	for i, breakpoint in ipairs(rss.playerState.hasteBreakpoints) do
		local targetHaste = cast.targetHaste[i];
		local deltaTicks = 0.0;
		local cache = { };
		local current = cast.events:Start();
		while current do
			local event = current.event;
			local eventType = event.eventType;
			if event.healed > 0 then
				if eventType == rss.EVENTS.SPELL_HEAL or eventType == rss.EVENTS.SPELL_PERIODIC_HEAL then
					local spellId = event.spellId
					if spellId and rss.SPELLS[spellId] and rss.SPELLS[spellId].ticks and 
						(spellId ~= rss.RIPTIDE or eventType == rss.EVENTS.SPELL_PERIODIC_HEAL) then
						if not cache[spellId] then
							local realTicks = rss.CalcTicksFromHaste(spellId, castHaste);
							local hypotheticalTicks = rss.CalcTicksFromHaste(spellId, targetHaste);
							cache[spellId] = (hypotheticalTicks/realTicks - 1);
						end
						deltaTicks = deltaTicks + event.healed * cache[spellId];
						if event.copiedBy then
							local copy = event.copiedBy:Start();
							while copy do
								local copyCast = copy.copier;
								if not copyCast.healing then
									rss.CastTotalHealing(copyCast);
								end
								if copyCast.originalHealing and copyCast.originalHealing > 0 then
									deltaTicks = deltaTicks + event.amount * cache[spellId] * (copyCast.healing/copyCast.originalHealing);
								end
								copy = copy.forward;
							end
						end
					end
				end
			end
			current = current.forward;
		end
		cast.deltaTicks[i] = deltaTicks;
	end
end

--[[
copyRaw = c*origRaw
f = copyEff/copyRaw
copyEff = f * copyRaw = f * c * origRaw
dcopyEff = f * c * dorigRaw = (copyEff/copyRaw) * (copyRaw/origRaw) * dorigRaw = (copyEff/origRaw) * dorigRaw
--]]

-- Estimate the change in overall healing, based on the reduction in cast time and activity, for all breakpoints
function rss.VaryCasts(cast)
	if cast.healing <= 0 or not cast.haste or not cast.busyFraction or cast.spellID == rss.ASCENDANCE or cast.spellID == rss.AURA_ANCESTRAL_GUIDANCE then return; end
	cast.deltaCasts = { };
	local castHaste = cast.haste;
	for i, breakpoint in ipairs(rss.playerState.hasteBreakpoints) do
		local fractionalDelta = cast.busyFraction * (cast.targetHaste[i] - castHaste) / (1 + castHaste);
		cast.deltaCasts[i] = cast.healing * fractionalDelta;
		if cast.copiedHealing then
			local copies = cast.copiedHealing:Start();
			while copies do
				local copiedTotal = copies.totalCopied;
				local copier = copies.cast;
				if not copier.healing then
					rss.CastTotalHealing(copier);
				end
				if copier.originalHealing and copier.originalHealing > 0 then
					cast.deltaCasts[i] = copiedTotal * (copier.healing/copier.originalHealing) * fractionalDelta;
				end
				copies = copies.forward;
			end
		end
	end
	local fractionalDelta = cast.busyFraction * cast.unitHaste / (1 + castHaste);
	cast.deltaCasts[0] = cast.healing * fractionalDelta;
	if cast.copiedHealing then
		local copies = cast.copiedHealing:Start();
		while copies do
			local copiedTotal = copies.totalCopied;
			local copier = copies.cast;
			if not copier.healing then
				rss.CastTotalHealing(copier);
			end
			if copier.originalHealing and copier.originalHealing > 0 then
				cast.deltaCasts[0] = copiedTotal * (copier.healing/copier.originalHealing) * fractionalDelta;
			end
			copies = copies.forward;
		end
	end
	
end

-- "banker's rounding" function to mirror Blizzard's calculations (round 0.5 to even)
function rss.bankRound(x)
	if x == 0 then return 0; end
	local signX = x/math.abs(x);
	local answer = signX * math.floor(math.abs(x + 0.5 * signX));
	if x - math.floor(x) == 0.5 then
		if answer/2 ~= math.floor(answer/2) then
			answer = answer - signX
		end
	end
	return answer;
end

-- determine the number of ticks and tick interval (in seconds) for a HoT or totem, given the player's *buffed* Spell Haste
-- from http://www.totemspot.com/vb/entry.php?b=55
function rss.CalcBuffedTicks(spellID)
	return rss.CalcTicksFromHaste(spellID, rss.playerState.buffedSpellHaste);
end

-- return the number of ticks and tick interval for a spell, given a specified haste% value
function rss.CalcTicksFromHaste(spellID, haste)
	local tickCount = 0;
	local tickInterval = 0;
	if rss.SPELLS[spellID] and rss.SPELLS[spellID].ticks then
		local ticks = rss.SPELLS[spellID].ticks;
		tickInterval = ticks.tick/(1 + haste);
		if rss.SPELLS[spellID].totem then
			tickCount = 1 + math.floor(ticks.duration/tickInterval);
		else
			tickCount = rss.bankRound(ticks.duration/tickInterval);
		end
	end
	return tickCount, tickInterval;
end

-- determine the spell haste% required to receive one more tick than current number
-- from http://www.totemspot.com/vb/entry.php?b=55
function rss.CalcBreakpoint(spellID, increment)
	if not rss.SPELLS[spellID] or not rss.SPELLS[spellID].ticks then return nil; end
	local ticks = rss.SPELLS[spellID].ticks;
	local currentTicks = rss.CalcBuffedTicks(spellID);
	if currentTicks < 1 then return; end
	local tickIncrement = increment;
	if not tickIncrement then tickIncrement = 1; end
	local newTicks = math.floor(currentTicks + tickIncrement);
	return rss.CalcHaste(spellID, newTicks);
end

-- Calculate the absolute spell haste value required for a spell to tick a specified number of times
-- This calculation is independent of the player's actual spell haste
function rss.CalcHaste(spellID, tickCount)
	if not rss.SPELLS[spellID] or not rss.SPELLS[spellID].ticks then return nil; end
	local ticks = rss.SPELLS[spellID].ticks;
	if tickCount <= rss.CalcTicksFromHaste(spellID, 0.0) then return 0.0; end
	local tickSpeed;
	if rss.SPELLS[spellID].totem then
		tickSpeed = (1000*ticks.duration)/(tickCount - 1); -- milliseconds
	else
		tickSpeed = (1000*ticks.duration)/(tickCount - 0.5); -- milliseconds
		if 2*math.floor(tickCount/2) == tickCount then
			tickSpeed = math.floor(tickSpeed) + 0.5;
		else
			tickSpeed = math.ceil(tickSpeed) - 0.5;
		end
	end
	return ((1000*ticks.tick)/tickSpeed - 1.0);
end

-- Calculate the player's minimum (no-gear) Haste value, based on talents, race and the always-assumed 5% raid buff
function rss.CalcMinHaste()
	return (1.05 * (rss.playerState.isGoblin and 1.01 or 1.0) * (rss.playerState.ancestralSwiftness and 1.05 or 1.0)) - 1.0;
end

-- Calculate the Haste Rating required for the spell to tick the specified number of times
-- The player is always assumed to have a 5% Haste buff
-- The Ancestral Swiftness talent buff and Goblin racial are applied if appropriate
function rss.CalcHasteRating(spellID, tickCount)
	local buffedHastePercent = math.max(0,rss.CalcHaste(spellID, tickCount));
	local unbuffedHastePercent = math.max(0,(1 + buffedHastePercent)/1.05 - 1);
	local hasteRatingPercent;
	if rss.playerState.isGoblin then
		unbuffedHastePercent = math.max(0, (1 + unbuffedHastePercent)/1.01 - 1);
	end
	if rss.playerState.ancestralSwiftnessActive then
		hasteRatingPercent = math.max(0,(1 + unbuffedHastePercent)/1.05 - 1);
	else
		hasteRatingPercent = unbuffedHastePercent;
	end
	local hasteRating = math.max(0,math.ceil(rss.RATING_TO_HASTE * hasteRatingPercent * 100));
	return hasteRating;
end

-- Calculate the ADDITIONAL Haste rating required for the spell to tick the specified number of times
-- The player is always assumed to have a +5% spell haste buff
-- The Ancestral Swiftness talent buff and goblin racial are applied if appropriate
function rss.CalcHasteRatingIncrement(spellID, tickCount)
	return math.ceil(rss.CalcHasteRating(spellID, tickCount)) - rss.playerState.baseHasteRating;
end

--[[
1+h_current = (1+race)*(1+talent)*(1+raid)*(1+hero)*(1+rating_current/c)
1+h_hyp = (1+race)*(1+talent)*(1+raid)*(1+hero)*(1+(rating_current + (rating_hyp-rating_base))/c)
1+h_hyp = (1+h_current)*(1+((rating_current-rating_base)+rating_hyp)/c)/(1+rating_current/c)

--]]
-- Calculate the buffed spell haste if the player's unbuffed haste rating were the specified valued
-- This includes all short- and long-term buffs: 
-- Goblin, Ancestral Swiftness, 5% raid buff, Heroism/Bloodlust, and any +Haste Rating buffs on top of the base rating value
-- It requires that the player's unbuffed Haste *Rating* is known
function rss.CalcBuffedHasteFromRating(rating)
	local currentRating = rss.playerState.hasteRating;
	local bonus = (rating+(currentRating-rss.playerState.baseHasteRating))/rss.RATING_TO_HASTE/100.0;
	local currentBonus = currentRating/rss.RATING_TO_HASTE/100.0;
	return (1+rss.playerState.buffedSpellHaste) * (1+bonus)/(1+currentBonus) - 1;

end

-- Does not depend on the player's gear, but does depend on race/talents
function rss.CalcHasteBreakpoints()
	local breakPoints = { };
	local minHaste = rss.CalcMinHaste();
	for spellID, spell in pairs(rss.SPELLS) do
		if spell.ticks and not spell.noHaste and spellID ~= rss.HEALING_STREAM_2T15 then
			local spellName = GetSpellInfo(spellID);
			if spellName then
				local tickCount = rss.CalcTicksFromHaste(spellID, minHaste) + 1;
				local hasteReq = rss.CalcHaste(spellID, tickCount);
				while hasteReq <= 0.5 do
					local ratingReq = rss.CalcHasteRating(spellID, tickCount);
					table.insert(breakPoints, { name = spellName , tickCount = tickCount , haste = hasteReq , rating = ratingReq });
					tickCount = tickCount + 1;
					hasteReq = rss.CalcHaste(spellID, tickCount);
				end
			end
		end
	end
	table.sort(breakPoints, function (a,b) if a.haste < b.haste then return true; end return false; end);
	return breakPoints;
end

function rss.PrintHasteIncrements()
	local current = { };
	for spellID, spell in pairs(rss.SPELLS) do
		if spell.ticks and not spell.noHaste and spellID ~= rss.HEALING_STREAM_2T15 then
			local spellName = GetSpellInfo(spellID);
			if spellName then
				current[spellID] = { name = spellName , tickCount = rss.CalcBuffedTicks(spellID) };
			end
		end
	end
	local increments = { };
	for spellID, currentInfo in pairs(current) do
		for delta = -1, 1 do
			local newTicks = currentInfo.tickCount + delta;
			local increment = rss.CalcHasteRatingIncrement(spellID, newTicks);
			if math.abs(increment) < rss.playerState.baseHasteRating then
				table.insert(increments, { name = currentInfo.name, tickCount = newTicks, increment = increment});
			end
		end
	end
	table.sort(increments, function (a,b) if a.increment < b.increment then return true; end return false; end);
	print("Haste breakpoints assuming +5% Haste raid buff and current talents:")
	for i, increment in ipairs(increments) do
		print(("[%d] %s %d ticks : %+d Haste Rating (%d total Haste Rating)"):format(i,increment.name,increment.tickCount,increment.increment, increment.increment + rss.playerState.baseHasteRating));
	end
end

function rss.PrintBreakpoints()
	for spellID, spell in pairs(rss.SPELLS) do
		if spell.ticks and not spell.noHaste and spellID ~= rss.HEALING_STREAM_2T15 then
			local spellName = GetSpellInfo(spellID);
			if spellName then
				local buffedBreakpointUp = 100*rss.CalcBreakpoint(spellID);
				local ratingIncrementUp = rss.CalcHasteRatingIncrement(spellID, math.floor(rss.CalcBuffedTicks(spellID)+1));
				print(("%s : %d ticks\nNeed %.2f%% buffed Haste or %.0f more Haste Rating for additional tick"):format(spellName, rss.CalcBuffedTicks(spellID), buffedBreakpointUp, ratingIncrementUp));
			end
		end
	end
end